# /polysquarecmakelinter/check_structure.py
#
# Linter checks for script structure
#
# See LICENCE.md for Copyright information
"""Linter checks for script structure"""

from cmakeast import ast_visitor
from polysquarecmakelinter import util
from polysquarecmakelinter.types import LinterFailure


def definitions_namespaced(contents, abstract_syntax_tree, **kwargs):
    """Checks that function and macro definitions are namespaced"""

    errors = []

    try:
        namespace = kwargs["namespace"]
    except KeyError:
        return errors

    def _definition_handler(name, node, depth):
        """Visits all definitions"""
        assert name == "FunctionDefinition" or name == "MacroDefinition"
        assert len(node.header.arguments) > 0

        del depth

        definition = node.header.arguments[0]
        def_name = definition.contents

        if not def_name.startswith((namespace, "_" + namespace)):
            msg = "Definition {0} does not start with {1}".format(def_name,
                                                                  namespace)
            replacement_name = "{0}_{1}".format(namespace, def_name)
            if def_name.startswith("_"):
                replacement_name = "_{0}".format(replacement_name)

            replacement = util.replace_word(contents[node.line - 1],
                                            definition.col - 1,
                                            def_name,
                                            replacement_name)

            errors.append(LinterFailure(msg, node.line, replacement))

    ast_visitor.recurse(abstract_syntax_tree,
                        function_def=_definition_handler,
                        macro_def=_definition_handler)

    return errors



# /polysquarecmakelinter/find_variables_in_scopes.py
#
# Find all variables and order by scope. Scoping in CMake is a bit strange,
# by default, all variables will have function scope, or failing that, global
# scope. Any variable set by foreach as a loop variable has scope only within
# that loop. Macro arguments have scope within their respective macros and
# anything set inside of a foreach loop has scope inside of its
# enclosing function.
#
# These functions don't necessarily follow the same scoping structure that
# CMake uses internally - scoping in CMake is actually based on the
# runtime call tree, so if function A calls function B then whatever
# was visible in function A will also be visible in function B. However,
# setting that variable will cause it to be set in function B's scope.
#
# Instead, the model followed is "logical scope" - variables from function
# B called from function B are not visible in function B, except if function
# B was defined inside function A. The same rules apply for macros.
#
# In the author's opinion, checks that rely on CMake's own scoping behaviour
# are really enforcing bad behaviour - implicit variable visibility is bad
# for maintainability and shouldn't be relied on in any case.
#
# Global Scope:
# set (VARIABLE VALUE)
#
# macro (my_macro MACRO_ARG)
#     MACRO_ARG has my_macro scope
#     set (MY_MACRO_VAR VALUE) MY_MACRO_VAR has my_function scope because it
#     was called by my_function
# endmacro ()
#
# function (my_other_function)
#     set (PARENT_VAR VALUE PARENT_SCOPE) PARENT_VAR has my_function scope
# endfunction ()
#
# function (my_function ARGUMENT)
#     ARGUMENT has my_function scope
#     set (MY_FUNC_VAR VALUE) MY_FUNC_VAR has my_function scope
#     foreach (LOOP_VAR ${MY_FUNC_VAR})
#         LOOP_VAR has foreach scope
#         set (VAR_SET_IN_LOOP VALUE) VAR_SET_IN_LOOP has my_function scope
#     endforeach ()
#     my_other_function ()
#     my_macro ()
# endfunction ()
#
# See LICENCE.md for Copyright information
"""Find all set variables and order by scope"""

import re

from collections import namedtuple
from polysquarecmakelinter import find_set_variables

Variable = namedtuple("Variable", "node source")
ScopeInfo = namedtuple("ScopeInfo", "name type")

_RE_VARIABLE_USE = re.compile(r"(?<![^\${])[0-9A-Za-z_]+(?![^]}])")

FOREACH_KEYWORDS = [
    "IN",
    "LISTS",
    "ITEMS"
]

IF_KEYWORDS = [
    "NOT",
    "STREQUAL",
    "STRLESS",
    "STRGREATER",
    "LESS",
    "EQUAL",
    "GREATER",
    "VERSION_LESS",
    "VERSION_EQUAL",
    "VERSION_GREATER",
    "EXISTS",
    "COMMAND",
    "POLICY",
    "TARGET",
    "IS_NEWER_THAN",
    "IS_DIRECTORY",
    "IS_ABSOLUTE",
    "MATCHES",
    "DEFINED",
    "AND",
    "OR",
    "OFF",
    "ON",
    "TRUE",
    "FALSE",
    "YES",
    "Y",
    "NO",
    "N",
    "IGNORE",
    "NOTFOUND"
]


# We use a class with constant variables here so that we can get int->int
# comparison. Comparing enums is slow because of the type lookup.
class VariableSource(object):
    """The source of a variable in a scope"""

    ForeachVar = 0
    MacroArg = 1
    FunctionArg = 2
    FunctionVar = 3
    MacroVar = 4
    GlobalVar = 5


class ScopeType(object):
    """The source of a variable in a scope"""

    Foreach = 0
    Macro = 1
    Function = 2
    Global = 3


class _Scope(object):
    """A place where variables are hoisted"""

    def __init__(self, info, parent):
        """Initializes parent"""
        super(_Scope, self).__init__()

        self.parent = parent
        self.info = info
        self.scopes = []

    def add_subscope(self, name, node, parent, factory):
        """Adds a new subscope"""

        assert node.__class__.__name__ != "ToplevelBody"

        node_scope_types = {
            "ForeachStatement": ScopeType.Foreach,
            "MacroDefinition": ScopeType.Macro,
            "FunctionDefinition": ScopeType.Function
        }

        assert node.__class__.__name__ in node_scope_types
        scope_type = node_scope_types[node.__class__.__name__]

        self.scopes.append(factory(ScopeInfo(name,
                                             scope_type),
                                   parent))


def traverse_scopes(abstract_syntax_tree,
                    body_function_call,
                    header_function_call,
                    factory):
    """Finds all set variables in tree and orders into scopes"""

    global_scope = factory(ScopeInfo("toplevel", ScopeType.Global), None)

    def _header_body_visitor(enclosing_scope, header_scope, header, body):
        """Visits a header-body like node"""
        if header is not None:

            header_function_call(header, enclosing_scope, header_scope)

        for statement in body:
            _node_recurse(statement, enclosing_scope, header_scope)

    def _node_recurse(node, enclosing_scope, header_scope):
        """Visits any node, adjusts scopes"""

        def _handle_if_block(node):
            """Handles if blocks"""

            _header_body_visitor(enclosing_scope,
                                 header_scope,
                                 node.if_statement.header,
                                 node.if_statement.body)

            for elseif in node.elseif_statements:
                _header_body_visitor(enclosing_scope,
                                     header_scope,
                                     elseif.header,
                                     elseif.body)

            if node.else_statement:
                _header_body_visitor(enclosing_scope,
                                     header_scope,
                                     node.else_statement.header,
                                     node.else_statement.body)

        def _handle_foreach_statement(node):
            """Handles foreach statements"""

            header_scope.add_subscope("foreach", node, header_scope, factory)

            _header_body_visitor(enclosing_scope,
                                 header_scope.scopes[-1],
                                 node.header,
                                 node.body)

        def _handle_while_statement(node):
            """Handles while statements"""

            _header_body_visitor(enclosing_scope,
                                 header_scope,
                                 node.header,
                                 node.body)

        def _handle_function_declaration(node):
            """Handles function declarations"""

            header_scope.add_subscope(node.header.arguments[0].contents,
                                      node,
                                      header_scope,
                                      factory)

            _header_body_visitor(header_scope.scopes[-1],
                                 header_scope.scopes[-1],
                                 node.header,
                                 node.body)

        def _handle_macro_declaration(node):
            """Handles macro declarations"""

            header_scope.add_subscope(node.header.arguments[0].contents,
                                      node,
                                      header_scope,
                                      factory)

            _header_body_visitor(header_scope.scopes[-1],
                                 header_scope.scopes[-1],
                                 node.header,
                                 node.body)

        def _handle_function_call(node):
            """Handles function calls - does nothing, nothing to recurse"""
            body_function_call(node, enclosing_scope, header_scope)

        def _handle_toplevel_body(node):
            """Handles the special toplevel body node"""

            _header_body_visitor(enclosing_scope,
                                 header_scope,
                                 None,
                                 node.statements)

        node_dispatch = {
            "IfBlock": _handle_if_block,
            "ForeachStatement": _handle_foreach_statement,
            "WhileStatement": _handle_while_statement,
            "FunctionDefinition": _handle_function_declaration,
            "MacroDefinition": _handle_macro_declaration,
            "ToplevelBody": _handle_toplevel_body,
            "FunctionCall": _handle_function_call
        }

        node_dispatch[node.__class__.__name__](node)

    _node_recurse(abstract_syntax_tree, global_scope, global_scope)

    return global_scope


def set_in_tree(abstract_syntax_tree):
    """Find variables set by scopes"""

    def scope_factory(info, parent):
        """Constructs a "set variables" scope"""
        class SetVariablesScope(_Scope):
            """Set variables in this scope"""

            def __init__(self, info, parent):
                """Initialzes set_vars member"""

                super(SetVariablesScope, self).__init__(info, parent)
                self.set_vars = []

        return SetVariablesScope(info, parent)

    def body_function_call(node, enclosing, body_header):
        """Handles function calls in a body and provides scope"""

        del body_header

        var_types = {
            ScopeType.Macro: VariableSource.MacroVar,
            ScopeType.Function: VariableSource.FunctionVar,
            ScopeType.Global: VariableSource.GlobalVar
        }

        set_var = find_set_variables.by_function_call(node)
        if set_var:

            # Special case for "set" and PARENT_SCOPE/CACHE scope
            if node.name == "set":
                try:
                    if node.arguments[2].contents == "PARENT_SCOPE":
                        assert enclosing.parent is not None
                        enclosing = enclosing.parent
                    elif node.arguments[2].contents == "CACHE":
                        while enclosing.parent is not None:
                            enclosing = enclosing.parent
                except IndexError:
                    pass
            # Another special case for set_property with GLOBAL as the
            # first argument. Create a notional "variable"
            elif node.name == "set_property":
                assert len(node.arguments[0]) >= 3
                if node.arguments[0].contents == "GLOBAL":
                    while enclosing.parent is not None:
                        enclosing = enclosing.parent

            info = enclosing.info
            enclosing.set_vars.append(Variable(set_var, var_types[info.type]))

    def header_function_call(node, header_enclosing, header):
        """Handles the "header" function call and provides scope"""

        del header_enclosing

        def _get_header_vars(header):
            """Adds variables implicitly set by header function call"""
            if header.name == "foreach":
                return [header.arguments[0]]
            elif header.name == "function" or header.name == "macro":
                return header.arguments[1:]
            else:
                return []

        var_types = {
            ScopeType.Foreach: VariableSource.ForeachVar,
            ScopeType.Function: VariableSource.FunctionArg,
            ScopeType.Macro: VariableSource.MacroArg
        }

        nodes = _get_header_vars(node)
        info = header.info
        header.set_vars += [Variable(v, var_types[info.type]) for v in nodes]

    return traverse_scopes(abstract_syntax_tree,
                           body_function_call,
                           header_function_call,
                           scope_factory)


def used_in_tree(abstract_syntax_tree):
    """Find variables used in scopes"""

    def scope_factory(info, parent):
        """Constructs a "set variables" scope"""
        class UsedVariablesScope(_Scope):
            """Used variables in this scope"""

            def __init__(self, info, parent):
                """Initialzes used_vars member"""

                super(UsedVariablesScope, self).__init__(info, parent)
                self.used_vars = []

        return UsedVariablesScope(info, parent)

    def body_function_call(node, body_enclosing, header):
        """Handles function calls in a node body"""
        del body_enclosing

        var_types = {
            ScopeType.Foreach: VariableSource.ForeachVar,
            ScopeType.Function: VariableSource.FunctionVar,
            ScopeType.Macro: VariableSource.MacroVar,
            ScopeType.Global: VariableSource.GlobalVar
        }

        for argument in node.arguments:
            if _RE_VARIABLE_USE.search(argument.contents) is not None:
                info = header.info
                header.used_vars.append(Variable(argument,
                                                 var_types[info.type]))

    def header_function_call(node, header_enclosing, current_header):
        """Handles function calls in a node header"""
        del header_enclosing

        var_types = {
            ScopeType.Foreach: lambda p: var_types[p.info.type](p.parent),
            ScopeType.Function: lambda _: VariableSource.FunctionVar,
            ScopeType.Macro: lambda _: VariableSource.MacroVar,
            ScopeType.Global: lambda _: VariableSource.GlobalVar
        }

        starts_new_header = ["foreach", "function", "macro"]
        if node.name in starts_new_header:
            header = current_header.parent
        else:
            header = current_header

        sct = header.info.type

        kw_exclude = {
            "if":  IF_KEYWORDS,
            "elseif": IF_KEYWORDS,
            "while": IF_KEYWORDS,
            "foreach": FOREACH_KEYWORDS,
            "function": [],
            "macro": [],
            "else": []
        }

        header_pos_exclude = {
            "if": lambda _: False,
            "elseif": lambda _: False,
            "while": lambda _: False,
            "foreach": lambda n: n == 0,
            "function": lambda _: True,
            "macro": lambda _: True,
            "else": lambda _: True
        }

        for index, argument in enumerate(node.arguments):
            is_var_use = _RE_VARIABLE_USE.search(argument.contents) is not None
            not_kw_excluded = argument.contents not in kw_exclude[node.name]
            not_pos_excluded = not header_pos_exclude[node.name](index)

            if is_var_use and not_kw_excluded and not_pos_excluded:
                variable_type = var_types[sct](header.parent)
                header.used_vars.append(Variable(argument, variable_type))

    return traverse_scopes(abstract_syntax_tree,
                           body_function_call,
                           header_function_call,
                           scope_factory)



# /polysquarecmakelinter/find_set_variables.py
#
# Details to find set variables in a module
#
# See LICENCE.md for Copyright information
"""Details to find set variables in a module"""

from collections import namedtuple
from cmakeast import ast_visitor
from polysquarecmakelinter import ignore


# Subclass namedtuple so that we can have a default argument for sub
class _SetVariable(namedtuple("_SetVariable", "cmd find sub")):
    """_SetVariable namedtuple with default variable for sub"""

    def __new__(cls, cmd, find, sub=None):
        """Factory with sub defaulted out"""
        return super(_SetVariable, cls).__new__(cls, cmd, find, sub)


def _find_arg_no(num):
    """Return an argument finder for argument at num"""
    class Finder(object):
        """Finder for arguments after num"""

        def find(self, arguments):  # pylint:disable=no-self-use
            """Returns argument at num"""
            try:
                return arguments[num]
            except IndexError:
                return None

        def generate(self,  # pylint:disable=no-self-use
                     sub,
                     other_transform,
                     match_transform):
            """Generates argument list of num size"""
            # A cheap hack to ensure that we generate a few extra in the
            # list GET case
            gen_num = abs(num - 1 if sub is not None else num)
            args_before = " ".join([other_transform("ARGUMENT")] * (gen_num))
            match_arg = match_transform("VALUE")
            subcommand = sub + " " if sub is not None else ""
            return subcommand + " ".join([args_before, match_arg])

    return Finder()


def _find_after_arg(argument):
    """Return an argument finder for an argument after argument"""
    class Finder(object):
        """Finder for arguments after argument"""

        def find(self, arguments):  # pylint:disable=no-self-use
            """Returns argument after argument or None if not found"""
            try:
                argument_index = 1
                for argument_index in range(1, len(arguments)):
                    if arguments[argument_index - 1].contents == argument:
                        break

                return arguments[argument_index]
            except IndexError:
                return None

        def generate(self,  # pylint:disable=no-self-use
                     sub,
                     other_transform,
                     match_transform):
            """Generates argument list of ARGUMENT__ {argument} VALUE"""
            args_before = other_transform("ARGUMENT__ {0}".format(argument))
            match_arg = match_transform("VALUE")
            subcommand = sub + " " if sub is not None else ""
            return subcommand + " ".join([args_before, match_arg])

    return Finder()


FUNCTIONS_SETTING_VARIABLES = [
    _SetVariable("aux_source_directory", _find_arg_no(1)),
    _SetVariable("build_command", _find_arg_no(0)),
    _SetVariable("cmake_host_system_information", _find_arg_no(1),
                 sub="RESULT"),
    _SetVariable("cmake_policy", _find_arg_no(2), sub="GET"),
    _SetVariable("execute_process", _find_after_arg("RESULT_VARIABLE")),
    _SetVariable("execute_process", _find_after_arg("OUTPUT_VARIABLE")),
    _SetVariable("execute_process", _find_after_arg("ERROR_VARIABLE")),
    _SetVariable("file", _find_arg_no(2), sub="READ"),
    _SetVariable("file", _find_arg_no(2), sub="STRINGS"),
    _SetVariable("file", _find_arg_no(2), sub="MD5"),
    _SetVariable("file", _find_arg_no(2), sub="SHA1"),
    _SetVariable("file", _find_arg_no(2), sub="SHA224"),
    _SetVariable("file", _find_arg_no(2), sub="SHA256"),
    _SetVariable("file", _find_arg_no(2), sub="SHA384"),
    _SetVariable("file", _find_arg_no(2), sub="SHA512"),
    _SetVariable("file", _find_arg_no(1), sub="GLOB"),
    _SetVariable("file", _find_arg_no(1), sub="GLOB_RECURSE"),
    _SetVariable("file", _find_arg_no(1), sub="RELATIVE_PATH"),
    _SetVariable("file", _find_arg_no(2), sub="TO_CMAKE_PATH"),
    _SetVariable("file", _find_arg_no(2), sub="TO_NATIVE_PATH"),
    _SetVariable("file", _find_after_arg("LOG"), sub="DOWNLOAD"),
    _SetVariable("file", _find_after_arg("LOG"), sub="UPLOAD"),
    _SetVariable("file", _find_after_arg("STATUS"), sub="DOWNLOAD"),
    _SetVariable("file", _find_after_arg("STATUS"), sub="UPLOAD"),
    _SetVariable("file", _find_arg_no(2), sub="TIMESTAMP"),
    _SetVariable("find_file", _find_arg_no(0)),
    _SetVariable("find_library", _find_arg_no(0)),
    _SetVariable("find_path", _find_arg_no(0)),
    _SetVariable("find_program", _find_arg_no(0)),
    _SetVariable("get_cmake_property", _find_arg_no(0)),
    _SetVariable("get_directory_property", _find_arg_no(0)),
    _SetVariable("get_filename_component", _find_arg_no(0)),
    _SetVariable("get_property", _find_arg_no(0)),
    _SetVariable("get_source_file_property", _find_arg_no(1)),
    _SetVariable("get_target_property", _find_arg_no(0)),
    _SetVariable("get_test_property", _find_arg_no(0)),
    _SetVariable("include", _find_after_arg("RESULT_VARIABLE")),
    _SetVariable("separate_arguments", _find_arg_no(1)),
    _SetVariable("set", _find_arg_no(0)),
    _SetVariable("unset", _find_arg_no(0)),
    _SetVariable("try_compile", _find_after_arg("RESULT_VAR")),
    _SetVariable("try_compile", _find_after_arg("OUTPUT_VARIABLE")),
    _SetVariable("try_run", _find_after_arg("RUN_RESULT_VAR")),
    _SetVariable("try_run", _find_after_arg("COMPILE_RESULT_VAR")),
    _SetVariable("try_run", _find_after_arg("RUN_OUTPUT_VARIABLE")),
    _SetVariable("try_run", _find_after_arg("OUTPUT_VARIABLE")),
    _SetVariable("list", _find_arg_no(2), sub="LENGTH"),
    _SetVariable("list", _find_arg_no(-1), sub="GET"),
    _SetVariable("list", _find_arg_no(3), sub="FIND"),
    _SetVariable("list", _find_arg_no(2), sub="REMOVE_ITEM"),
    _SetVariable("list", _find_arg_no(1), sub="REMOVE_AT"),
    _SetVariable("list", _find_arg_no(1), sub="REMOVE_DUPLICATES"),
    _SetVariable("list", _find_arg_no(1), sub="REVERSE"),
    _SetVariable("list", _find_arg_no(1), sub="SORT"),
    _SetVariable("list", _find_arg_no(1), sub="APPEND"),
    _SetVariable("list", _find_arg_no(1), sub="INSERT"),
    _SetVariable("match", _find_arg_no(1), sub="EXPR")
]

_FUNCTIONS_SETTING_VARS_INT = (FUNCTIONS_SETTING_VARIABLES +
                               [_SetVariable("set_property",
                                             _find_after_arg("PROPERTY"),
                                             sub="GLOBAL")])


def by_function_call(node):
    """Returns a variable (as a word node) set by a function call, if any"""
    for matcher in _FUNCTIONS_SETTING_VARS_INT:
        if matcher.cmd == node.name:
            # Exclude where subcommand does not match
            # Note: pylint trips up on the generated code and believes that
            # matcher has no attribute named "sub". This is incorrect, so
            # a warning is suppressed here.
            if matcher.sub is not None:  # pylint:disable=no-member
                try:
                    if (matcher.sub !=  # pylint:disable=no-member
                            node.arguments[0].contents):
                        continue
                except IndexError:
                    continue

            evaluate = matcher.find.find(node.arguments)

            if evaluate:
                return evaluate

    return None


def in_tree(abstract_syntax_tree):
    """Find all assigned variables in this abstract_syntax_tree"""

    variables = []

    def _call_visitor(name, node):
        """Visits all function calls"""
        assert name == "FunctionCall"

        evaluate = by_function_call(node)
        if evaluate:
            variables.append(evaluate)

    ast_visitor.recurse(abstract_syntax_tree,
                        function_call=ignore.visitor_depth(_call_visitor))

    return variables



# /polysquarecmakelinter/ignore.py
#
# Wrapper functions to ignore certain arguments in callbacks
#
# See LICENCE.md for Copyright information
"""Wrapper functions to ignore certain arguments in callbacks"""


def all_but_ast(check):
    """Only passes AST to check"""

    def _check_wrapper(contents, ast, **kwargs):
        """Wraps check and passes the AST to it"""

        del contents
        del kwargs

        return check(ast)

    return _check_wrapper


def check_kwargs(check):
    """Does not pass kwargs to check"""

    def _check_wrapper(contents, ast, **kwargs):
        """Does not pass kwargs to check"""

        del kwargs

        return check(contents, ast)

    return _check_wrapper


def visitor_depth(visitor):
    """Does not pass depth to visitor"""

    def _visitor_wrapper(name, node, depth):
        """Does not pass depth to visitor"""

        del depth

        visitor(name, node)

    return _visitor_wrapper



# /polysquarecmakelinter/check_style.py
#
# Linter checks for style guide
#
# See LICENCE.md for Copyright information
"""Linter checks for style guide"""

import re

from cmakeast import ast_visitor
from cmakeast.ast import WordType
from polysquarecmakelinter import find_set_variables
from polysquarecmakelinter.types import LinterFailure
from polysquarecmakelinter import ignore
from polysquarecmakelinter import util

_RE_DOUBLE_OUTER_QUOTES = re.compile("^\".*\"$")
_RE_IF_CONDITION = re.compile(r"if|elseif")
_RE_IS_DEFINITION = re.compile(r"function|macro")

_RE_HEADERBODY_LIKE = re.compile(r"WhileStatement|"
                                 "ForeachStatement|"
                                 "FunctionDefinition|"
                                 "MacroDefinition")
_RE_FUNCTION_CALL = re.compile(r"FunctionCall")
_RE_IF_BLOCK = re.compile(r"IfBlock")
_RE_TOPLEVEL = re.compile(r"ToplevelBody")


def space_before_call(contents, abstract_syntax_tree):
    """Checks that each function call is preceded by a single space"""

    errors = []

    def _call_handler(name, node):
        """Handles function calls"""

        assert name == "FunctionCall"

        line_index = node.line - 1
        col_index = node.col - 1
        end_of_name_index = col_index + len(node.name)
        num_spaces_until_openparen = 0
        index = end_of_name_index
        while contents[line_index][index] != "(":
            index += 1

        # AST checks will assert if the open paren is not found here

        num_spaces_until_openparen = index - end_of_name_index

        # Must be one space only
        if num_spaces_until_openparen != 1:
            extra_spaces = num_spaces_until_openparen - 1
            msg = "{0} extra spaces between call of {1}".format(extra_spaces,
                                                                node.name)
            replacement = util.replace_word(contents[line_index],
                                            end_of_name_index,
                                            " " * (extra_spaces + 1),
                                            " ")
            errors.append(LinterFailure(msg, node.line, replacement))

    ast_visitor.recurse(abstract_syntax_tree,
                        function_call=ignore.visitor_depth(_call_handler))

    return errors


def lowercase_functions(contents, abstract_syntax_tree):
    """Checks that function / macro usage is all lowercase"""

    errors = []

    def _call_handler(name, node):
        """Handles function calls"""
        assert name == "FunctionCall"

        if node.name.lower() != node.name:
            msg = "{0} is not lowercase".format(node.name)
            replacement = util.replace_word(contents[node.line - 1],
                                            node.col - 1,
                                            node.name,
                                            node.name.lower())
            errors.append(LinterFailure(msg, node.line, replacement))

    def _definition_handler(name, node):
        """Handles function calls and macro definitions"""
        assert name == "FunctionDefinition" or name == "MacroDefinition"

        definition_name_word = node.header.arguments[0]
        definition_name = definition_name_word.contents

        if definition_name.lower() != definition_name:
            msg = "{0} name {1} is not lowercase".format(name, definition_name)
            line_index = definition_name_word.line - 1
            col_index = definition_name_word.col - 1
            replacement = util.replace_word(contents[line_index],
                                            col_index,
                                            definition_name,
                                            definition_name.lower())
            errors.append(LinterFailure(msg,
                                        definition_name_word.line,
                                        replacement))

    ast_visitor.recurse(abstract_syntax_tree,
                        function_call=ignore.visitor_depth(_call_handler),
                        function_def=ignore.visitor_depth(_definition_handler),
                        macro_def=ignore.visitor_depth(_definition_handler))

    return errors


def uppercase_arguments(contents, abstract_syntax_tree):
    """Checks that arguments to definitions are all uppercase"""

    errors = []

    def _definition_visitor(name, node):
        """Visit all definitions"""

        assert name == "FunctionDefinition" or name == "MacroDefinition"

        if len(node.header.arguments) < 2:
            return

        for arg in node.header.arguments[1:]:
            if arg.contents.upper() != arg.contents:
                msg = "{0} must be uppercase".format(arg.contents)
                line_index = arg.line - 1
                col_index = arg.col - 1
                replacement = util.replace_word(contents[line_index],
                                                col_index,
                                                arg.contents,
                                                arg.contents.upper())
                errors.append(LinterFailure(msg, arg.line, replacement))

    ast_visitor.recurse(abstract_syntax_tree,
                        function_def=ignore.visitor_depth(_definition_visitor),
                        macro_def=ignore.visitor_depth(_definition_visitor))

    return errors


def set_variables_capitalized(contents, abstract_syntax_tree):
    """Checks that each variable mutated is capitalized"""

    errors = []

    variables = find_set_variables.in_tree(abstract_syntax_tree)
    for evaluate in variables:

        evaluate_upper = evaluate.contents.upper()

        # Argument should be either String, VariableDereference
        # or Variable and a transformation to uppercase
        # should have no effect
        if not (util.is_word_sink_variable(evaluate.type) and
                evaluate_upper == evaluate.contents):
            desc = "{0} must be uppercase".format(evaluate.contents)
            line = evaluate.line - 1
            replacement = util.replace_word(contents[line],
                                            evaluate.col - 1,
                                            evaluate.contents,
                                            evaluate_upper)
            errors.append(LinterFailure(desc,
                                        evaluate.line,
                                        replacement))
            break

    return errors


def func_args_aligned(contents, abstract_syntax_tree):
    """Checks that function arguments are aligned


    Function arguments must be aligned either to the same line or
    must fall on the same column as the last argument on the line of
    the function call. The only special case is for definitions, where we
    align after the second argument, (eg the first argument to the defined
    function or macro)
    """

    errors = []

    def _check_horizontal_space(node, index):
        """Checks horizontal space between arguments on same line"""
        if index > 0:
            current_column = node.arguments[index].col
            previous_column = node.arguments[index - 1].col
            previous_len = len(node.arguments[index - 1].contents)
            spaces_start = previous_column + previous_len
            num_spaces = current_column - (previous_column + previous_len)

            line_number = node.arguments[index].line - 1

            if num_spaces != 1:
                cur = node.arguments[index].contents
                prev = node.arguments[index - 1].contents

                msg = "Must be a single space between {0} and {1}".format(cur,
                                                                          prev)
                replacement = util.replace_word(contents[line_number],
                                                spaces_start,
                                                " " * num_spaces,
                                                " ")
                errors.append(LinterFailure(msg, node.line, replacement))

    def _call_visitor(name, node):
        """Visits all function calls"""
        assert name == "FunctionCall"

        class AlignmentInfo(object):
            """Mutable alignment info"""

            def __init__(self):
                """Initialize"""

                super(AlignmentInfo, self).__init__()

                self.col = None
                self.line = None
                self.line_has_kw = False

        align = AlignmentInfo()
        arguments_len = len(node.arguments)

        if arguments_len == 0:
            return

        # Check if this was a defintion. If so, then we align after the last
        # argument and not after the first.
        is_definition = (_RE_IS_DEFINITION.match(node.name) is not None)
        is_if_cond = (_RE_IF_CONDITION.match(node.name) is not None)

        if is_definition and arguments_len > 1:
            baseline_col = node.arguments[1].col
        else:
            baseline_col = node.arguments[0].col

        align.col = node.arguments[0].col
        align.line = node.arguments[0].line

        def _check_alignment(arg, index):
            """Checks alignment of arg"""

            if align.line == arg.line:
                _check_horizontal_space(node, index)

            if arg.col == baseline_col:
                # Baseline column - reset align.col

                if (arg.type == WordType.Variable or
                        is_if_cond or
                        arg.line == node.arguments[0].line):
                    # This might be a keyword arg - allow alignment to this
                    # line. Otherwise it is the first argument on this line,
                    # or a function call where we allow arguments on any line
                    align.line = arg.line
                else:
                    # This is not a keyword arg. Do not allow alignment to
                    # this line (only this column)
                    align.line = None

                if arg.type == WordType.Variable:
                    align.line_has_kw = True
                else:
                    align.line_has_kw = False
                align.col = arg.col

            # If we're on our currently aligned line, and this line starts
            # with a keyword-like argument, then allow align.col
            # to shift appropriately, check horizontal space
            if align.line == arg.line and align.line_has_kw:
                align.col = arg.col

            misaligned_col = align.col and align.col != arg.col
            misaligned_line = align.line and align.line != arg.line

            if ((misaligned_col or not align.col) and
                    (misaligned_line or not align.line)):

                msg_parts = []
                replacement = None

                if misaligned_line is not None:
                    msg_parts.append("line {0}".format(align.line))

                if misaligned_col is not None:
                    msg_parts.append("col {0}".format(align.col))

                if align.col != baseline_col:
                    msg_parts.append("col {0}".format(baseline_col))

                if misaligned_line is None:
                    line_idx = arg.line - 1
                    offset = min(0, align.col - arg.col)
                    spaces = max(0, align.col - arg.col)
                    replacement = util.replace_word(contents[line_idx],
                                                    arg.col + offset,
                                                    " " * offset,
                                                    " " * spaces)

                msg = ("Argument {0} must fall on any of: "
                       "{1}".format(arg.contents,
                                    ", ".join(msg_parts)))
                errors.append(LinterFailure(msg,
                                            arg.line,
                                            replacement))

        for index in range(0, len(node.arguments)):
            arg = node.arguments[index]
            # If the argument is on the same line as the function call, then
            # update align.col and make sure that the argument is one
            # space away fromt the last argument
            _check_alignment(arg, index)

    ast_visitor.recurse(abstract_syntax_tree,
                        function_call=ignore.visitor_depth(_call_visitor))
    return errors


def double_outer_quotes(contents, abstract_syntax_tree):
    """Checks that all outer quotes are double quotes"""

    errors = []

    def _word_visitor(name, node):
        """Visits all arguments"""
        assert name == "Word"

        if node.type == WordType.String:
            if not _RE_DOUBLE_OUTER_QUOTES.match(node.contents):
                msg = "{0} must use double quotes".format(node.contents)
                replacement_word = "\"{0}\"".format(node.contents[1:-1])
                replacement = util.replace_word(contents[node.line - 1],
                                                node.col - 1,
                                                node.contents,
                                                replacement_word)
                errors.append(LinterFailure(msg, node.line, replacement))

    ast_visitor.recurse(abstract_syntax_tree,
                        word=ignore.visitor_depth(_word_visitor))

    return errors


def calls_indented_correctly(contents, abstract_syntax_tree, **kwargs):
    """Check that all calls to functions are indented at the correct level"""

    errors = []

    try:
        indent = kwargs["indent"]
    except KeyError:
        return errors

    def _visit_function_call(node, depth):
        """Hanldes function calls"""

        col = node.col
        expected = 1 + (depth * indent)
        if node.col != expected:
            delta = expected - node.col
            msg = "Expected {0} to be on column {1}".format(node.name,
                                                            expected)
            replacement = util.replace_word(contents[node.line - 1],
                                            col - 1 + min(0, delta),
                                            " " * max(0, delta * -1),
                                            " " * max(0, delta))
            errors.append(LinterFailure(msg, node.line, replacement))

    def _visit_header_body_block(node, depth):
        """Handles header/body like statements"""

        _visit_function_call(node.header, depth)

        for sub in node.body:
            _node_dispatch(sub, depth + 1)

        if getattr(node, "footer", None) is not None:
            _visit_function_call(node.footer, depth)

    def _visit_if_block(node, depth):
        """Handles if blocks"""

        _visit_header_body_block(node.if_statement, depth)

        for elseif_statement in node.elseif_statements:
            _visit_header_body_block(elseif_statement, depth)

        if node.else_statement:
            _visit_header_body_block(node.else_statement, depth)

        _visit_function_call(node.footer, depth)

    def _visit_toplevel(node, depth):
        """Handles toplevel blocks"""
        assert depth == 0

        for statement in node.statements:
            _node_dispatch(statement, depth)

    indent_node_dispatch = [
        (_RE_HEADERBODY_LIKE, _visit_header_body_block),
        (_RE_IF_BLOCK, _visit_if_block),
        (_RE_FUNCTION_CALL, _visit_function_call),
        (_RE_TOPLEVEL, _visit_toplevel)
    ]

    def _node_dispatch(node, depth):
        """Dispatches per-node, handling depth like actual indent depth"""

        for regex, handler in indent_node_dispatch:
            if regex.match(node.__class__.__name__):
                handler(node, depth)

    _node_dispatch(abstract_syntax_tree, 0)

    return errors



# /polysquarecmakelinter/check_access.py
#
# Linter checks for access rights
#
# See LICENCE.md for Copyright information
"""Linter checks for access rights"""

from polysquarecmakelinter.types import LinterFailure
from polysquarecmakelinter import find_all
from polysquarecmakelinter import find_variables_in_scopes


def only_use_own_privates(abstract_syntax_tree):
    """Checks that all private definitions used are defined here"""

    calls, defs = find_all.private_calls_and_definitions(abstract_syntax_tree)

    errors = []

    for call, info in calls.items():
        if call not in defs.keys():
            for line in info:
                msg = "Used external private definition {0}".format(call)
                errors.append(LinterFailure(msg, line))

    return errors


def only_use_own_priv_vars(ast):
    """Checks that all private variables used are defined here"""

    errors = []

    global_set_vars = find_variables_in_scopes.set_in_tree(ast)
    global_used_vars = find_variables_in_scopes.used_in_tree(ast)

    # The big assumption here is that the "scopes" structure in both
    # trees are the same
    def _scope_visitor(set_vars_scope, used_vars_scope):
        """Visits scope's set vars and used vars.


        If a var was private and used, but not set in this scope or any
        parents, then report an error
        """

        assert len(set_vars_scope.scopes) == len(used_vars_scope.scopes)

        for index in range(0, len(set_vars_scope.scopes)):
            _scope_visitor(set_vars_scope.scopes[index],
                           used_vars_scope.scopes[index])

        for variable in used_vars_scope.used_vars:
            for use in find_all.variables_used_in_expr(variable.node):
                if use.startswith("_"):
                    # Used a private, check if it was set
                    private_var_was_set = False

                    traverse_scope = set_vars_scope
                    while traverse_scope is not None:
                        for set_var in traverse_scope.set_vars:
                            if (set_var.node.contents == use and
                                    (set_var.node.line,
                                     set_var.node.col) != variable):
                                private_var_was_set = True
                                break

                        traverse_scope = traverse_scope.parent

                    if not private_var_was_set:
                        msg = ("Referenced external private"
                               " variable {0}").format(use)
                        errors.append(LinterFailure(msg,
                                                    variable.node.line,
                                                    variable.node.col))

    _scope_visitor(global_set_vars, global_used_vars)

    return errors



# /polysquarecmakelinter/check_unused.py
#
# Linter checks for unused definitions
#
# See LICENCE.md for Copyright information
"""Linter checks for Linter checks for unused definitions"""

import re

from cmakeast.ast import WordType
from polysquarecmakelinter.types import LinterFailure
from polysquarecmakelinter import (find_variables_in_scopes,
                                   find_all)

_RE_VARIABLE_USE = re.compile(r"(?<![^\${])[0-9A-Za-z_]+(?![^]}])")


def vars_in_func_used(abstract_syntax_tree):
    """Checks that variables defined in a function are used later"""

    errors = []

    set_scopes = find_variables_in_scopes.set_in_tree(abstract_syntax_tree)
    used_scopes = find_variables_in_scopes.used_in_tree(abstract_syntax_tree)

    # Iterate through the set and used variables - making sure that any set
    # variables are used somewhere down the scope chain
    def _scope_visitor(set_scope, used_scope):
        """Visits both scopes at the same level"""

        assert len(set_scope.scopes) == len(used_scope.scopes)

        def _used_scope_finder(used_scope, node):
            """Searches for uses of variable_node"""

            def _is_candidate_use(use, node):
                """Checks if use_node is not node, but also a candidate"""
                if use.node == node:
                    return False

                used_vars_in_node = _RE_VARIABLE_USE.findall(use.node.contents)
                for used_var in used_vars_in_node:
                    if used_var == node.contents:
                        return True

            used_vs = used_scope.used_vars
            matched_uses = [u for u in used_vs if _is_candidate_use(u, node)]

            # Found use, done
            if len(matched_uses):
                return True

            # Didn't find use - keep going down if possible:
            for subscope in used_scope.scopes:
                if _used_scope_finder(subscope, node):
                    return True

            return False

        # Ignore the global scope
        if id(set_scopes) != id(set_scope):
            for var in set_scope.set_vars:

                if var.node.type != WordType.Variable:
                    return

                if not _used_scope_finder(used_scope, var.node):
                    msg = "Unused local variable {0}".format(var.node.contents)
                    errors.append(LinterFailure(msg, var.node.line))

        for index in range(0, len(set_scope.scopes)):
            _scope_visitor(set_scope.scopes[index],
                           used_scope.scopes[index])

    _scope_visitor(set_scopes, used_scopes)

    return errors


def private_vars_at_toplevel(abstract_syntax_tree):
    """Checks that private variables defined at the top level are used later"""

    errors = []

    variables_set = find_all.toplevel_set_private_vars(abstract_syntax_tree)

    def _not_in_variables_set(node):
        """Returns false if in variables_set"""

        try:
            return not (node.line, node.col) in variables_set[node.contents]
        except KeyError:
            return True

    def _starts_with_underscore(name):
        """Returns true if variable name starts with _"""

        return name.startswith("_")

    variables_used = find_all.variables_used_matching(abstract_syntax_tree,
                                                      _not_in_variables_set,
                                                      _starts_with_underscore)

    # Check the intersection
    for var in variables_set.keys():
        if var not in variables_used.keys():
            msg = "Unused set variable at toplevel {0}".format(var)
            errors.append(LinterFailure(msg, variables_set[var][0][0]))

    return errors


def private_definitions_used(abstract_syntax_tree):
    """Checks that all private definitions are used by this module"""

    calls, defs = find_all.private_calls_and_definitions(abstract_syntax_tree)

    errors = []

    for definition, info in defs.items():
        if definition not in calls.keys():
            for line in info:
                msg = "Unused private definition {0}".format(definition)
                errors.append(LinterFailure(msg, line))

    return errors



# /polysquarecmakelinter/linter.py
#
# Entry point for linter.
#
# See LICENCE.md for Copyright information
""" Main module for linter """

from cmakeast import ast
from polysquarecmakelinter import check_access as access
from polysquarecmakelinter import check_correctness as correct
from polysquarecmakelinter import check_structure as structure
from polysquarecmakelinter import check_style as style
from polysquarecmakelinter import check_unused as unused
from polysquarecmakelinter import ignore

import argparse
import os
import re
import sys

_RE_NOLINT = re.compile(r"^.*#\s+NOLINT:")


def should_ignore(line, warning):
    """Specify whether or not to ignore warnings on this line"""
    match = _RE_NOLINT.search(line)
    if match:
        try:
            # Special case, "*" means "all errors"
            if line[match.end():][0] == "*":
                return True
            elif line[match.end():match.end() + len(warning)] == warning:
                return True
        except IndexError:
            pass

    return False


LINTER_FUNCTIONS = {
    "structure/namespace": structure.definitions_namespaced,
    "style/space_before_func": ignore.check_kwargs(style.space_before_call),
    "style/set_var_case": ignore.check_kwargs(style.set_variables_capitalized),
    "style/uppercase_args": ignore.check_kwargs(style.uppercase_arguments),
    "style/lowercase_func": ignore.check_kwargs(style.lowercase_functions),
    "style/argument_align": ignore.check_kwargs(style.func_args_aligned),
    "style/doublequotes": ignore.check_kwargs(style.double_outer_quotes),
    "style/indent": style.calls_indented_correctly,
    "correctness/quotes": ignore.check_kwargs(correct.path_variables_quoted),
    "unused/private": ignore.all_but_ast(unused.private_definitions_used),
    "unused/var_in_func": ignore.all_but_ast(unused.vars_in_func_used),
    "unused/private_var": ignore.all_but_ast(unused.private_vars_at_toplevel),
    "access/other_private": ignore.all_but_ast(access.only_use_own_privates),
    "access/private_var": ignore.all_but_ast(access.only_use_own_priv_vars)
}


def lint(contents,
         whitelist=None,
         blacklist=None,
         **kwargs):
    """Actually lints some file contents.

    Contents should be a raw string with \n. whitelist is a list of checks
    to only perform, blacklist is list of checks to never perform.
    """

    abstract_syntax_tree = ast.parse(contents)
    contents_lines = contents.splitlines(True)
    linter_functions = LINTER_FUNCTIONS

    def _keyvalue_pair_if(dictionary, condition):
        """Returns a key-value pair in dictionary if condition matched"""
        return {
            k: v for (k, v) in dictionary.items() if condition(k)
        }

    whitelist = whitelist if whitelist is not None else []
    blacklist = blacklist if blacklist is not None else []

    linter_functions = _keyvalue_pair_if(linter_functions,
                                         lambda k: k in whitelist)
    linter_functions = _keyvalue_pair_if(linter_functions,
                                         lambda k: k not in blacklist)

    linter_errors = []
    for (code, function) in linter_functions.items():
        errors = function(contents_lines,
                          abstract_syntax_tree,
                          **kwargs)
        for error in errors:
            linter_errors.append((code, error))

    return linter_errors


class ShowAvailableChecksAction(argparse.Action):
    """If --checks is encountered, just show available checks and exit"""

    def __call__(self, parser, namespace, values, option_string=None):

        del namespace
        del parser
        del values

        if option_string == "--checks":
            sys.stdout.write("Available option are:\n")
            for item in LINTER_FUNCTIONS.keys():
                sys.stdout.write(" * {0}\n".format(item))

            sys.exit(0)


def _parse_arguments():
    """Returns a parser context result"""

    parser = argparse.ArgumentParser(description="Lint for Polysquare "
                                     "style guide")
    parser.add_argument("--checks",
                        nargs=0,
                        action=ShowAvailableChecksAction,
                        help="list available checks")
    parser.add_argument("files",
                        nargs="*",
                        metavar=("FILE"),
                        help="read FILE",
                        type=argparse.FileType("r+"))
    parser.add_argument("--whitelist",
                        nargs="*",
                        help="list of checks that should only be run",
                        default=None)
    parser.add_argument("--blacklist",
                        nargs="*",
                        help="list of checks that should never be run",
                        default=None)
    parser.add_argument("--indent",
                        nargs=1,
                        type=int,
                        help="Indent level",
                        default=None)
    parser.add_argument("--namespace",
                        nargs=1,
                        type=str,
                        help="Namespace for functions",
                        default=None)
    parser.add_argument("--fix-what-you-can",
                        action="store_const",
                        const=True)

    return parser.parse_args()


def _report_lint_error(error, file_path):
    """Report a linter error"""
    line = error[1].line
    code = error[0]
    description = error[1].description
    sys.stderr.write("{0}:{1} [{2}] {3}".format(file_path,
                                                line,
                                                code,
                                                description))


def _apply_replacement(error, found_file, file_lines):
    """Apply a single replacement"""
    fixed_lines = file_lines
    fixed_lines[error[1].line - 1] = error[1].replacement
    concatenated_fixed_lines = "".join(fixed_lines)

    # Only fix one error at a time
    found_file.seek(0)
    found_file.write(concatenated_fixed_lines)
    found_file.truncate()


def main():
    """Entry point for the linter"""
    result = _parse_arguments()

    num_errors = 0
    for found_file in result.files:
        file_path = os.path.abspath(found_file.name)
        file_contents = found_file.read()
        file_lines = file_contents.splitlines(True)
        try:
            kwargs = {}
            if result.namespace is not None:
                kwargs["namespace"] = result.namespace[0]

            if result.indent is not None:
                kwargs["indent"] = result.indent[0]

            errors = lint(file_contents,  # pylint:disable=star-args
                          result.whitelist,
                          result.blacklist,
                          **kwargs)
        except RuntimeError as err:
            msg = "RuntimeError in processing {0} - {1}".format(file_path,
                                                                str(err))
            raise RuntimeError(msg)

        for error in errors:
            if not should_ignore(file_lines[error[1].line - 1], error[0]):
                _report_lint_error(error, file_path)
                if (result.fix_what_you_can and
                        error[1].replacement is not None):
                    _apply_replacement(error, found_file, file_lines)
                    sys.stderr.write(" ... FIXED\n")
                    break

                sys.stderr.write("\n")

                num_errors += 1

    sys.exit(num_errors)



# /polysquarecmakelinter/util.py
#
# Utility functions shared amongst checks
#
# See LICENCE.md for Copyright information
"""Utility functions shared amongst checks"""

from cmakeast.ast import WordType


def replace_word(line, start, word, replacement):
    """Helper function to replace a word starting at start with replacement"""
    return line[:start] + replacement + line[start + len(word):]


def is_word_sink_variable(word_type):
    """Returns true if this word can be used to set a value"""
    return word_type in [WordType.Variable, WordType.String]


def is_word_maybe_path(word_type):
    """Returns true if this word might be an unquoted path"""
    return word_type in [WordType.VariableDereference,
                         WordType.CompoundLiteral]
